from flymyai.core.clients.AsyncClient import BaseAsyncClient
from flymyai.core.clients.SyncClient import BaseSyncClient


class FlyMyAI(BaseSyncClient): ...


class AsyncFlyMyAI(BaseAsyncClient): ...



import dataclasses

import httpx


@dataclasses.dataclass
class ResponseLike:
    status_code: int
    url: httpx.URL
    content: bytes

    def to_msg(self):
        return f"""
            BAD REQUEST DETECTED ({self.status_code}):
            REQUEST URL: {self.url};
        """



import os
from typing import Optional, Callable, AsyncContextManager, Awaitable

import httpx

from flymyai.core.response_factory.plain_inference_response_factory import (
    SSEInferenceResponseFactory,
)
from flymyai.core._streaming import SSEDecoder
from flymyai.core.authorizations import APIKeyClientInfo
from flymyai.core.clients.base_client import BaseClient, _predict_timeout
from flymyai.core.exceptions import (
    BaseFlyMyAIException,
    FlyMyAIOpenAPIException,
    FlyMyAIPredictException,
    FlyMyAIExceptionGroup,
    FlyMyAIAsyncTaskException,
)
from flymyai.core.models.successful_responses import (
    OpenAPISchemaResponse,
    PredictionResponse,
    AsyncPredictionTask,
)
from flymyai.core.stream_iterators.AsyncPredictionStream import AsyncPredictionStream
from flymyai.multipart import MultipartPayload
from flymyai.utils.utils import aretryable_callback


class BaseAsyncClient(BaseClient[httpx.AsyncClient]):
    def _construct_client(self):
        return httpx.AsyncClient(
            http2=True,
            headers=self.client_info.authorization_headers,
            base_url=os.getenv("FLYMYAI_DSN", "https://api.flymy.ai/"),
            timeout=_predict_timeout,
        )

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self, "_client"):
            await self._client.aclose()

    async def openapi_schema(self, model: Optional[str] = None, max_retries=None):
        """
        :param max_retries: retries before giving up
        :return:
        :return: OpenAPISchemaResponse(exc_history, openapi_schema, response):
                exc_history - dict with exceptions;
                openapi_schema - dict with openapi;
        """
        history, response = await aretryable_callback(
            lambda: self._openapi_schema(),
            max_retries or self.max_retries,
            FlyMyAIPredictException,
            FlyMyAIExceptionGroup,
        )
        return OpenAPISchemaResponse.from_response(
            exc_history=history, openapi_schema=response.json(), response=response
        )

    def _openapi_schema(self, client_info: APIKeyClientInfo):
        """
        OpenAPI request for the current project, wrapped in executor-method (using HTTP/1)
        :return:
        """
        try:
            return self._wrap_request(
                lambda: self._client.get(
                    client_info.openapi_schema_path,
                    headers=client_info.authorization_headers,
                )
            )
        except BaseFlyMyAIException as e:
            raise FlyMyAIOpenAPIException.from_response(e.response)

    async def cancel_prediction(
        self,
        prediction_id: str,
        model: Optional[str] = None,
        client_info: APIKeyClientInfo = None,
    ):
        if client_info:
            full_client_info = client_info
        else:
            full_client_info = self.amend_client_info(model)
        response = await self._client.patch(
            url=full_client_info.prediction_cancel_path,
            json={"infer_id": prediction_id},
        )
        return SSEInferenceResponseFactory(
            httpx_response=response, httpx_request=response.request
        ).construct()

    @classmethod
    async def _sse_instant(
        cls, async_response_stream: Callable[[], AsyncContextManager[httpx.Response]]
    ):
        """
        A non-blocking approach to fetch a response stream
        :param async_response_stream: context manager with underlying stream
        :return: FlyMyAIResponse
        """
        async with async_response_stream() as stream:
            sse = await SSEDecoder().aiter(stream.aiter_lines()).__anext__()
            try:
                response = SSEInferenceResponseFactory(
                    sse=sse, httpx_request=stream.request, httpx_response=stream
                ).construct()
                return response
            except BaseFlyMyAIException as e:
                raise FlyMyAIPredictException.from_base_exception(e)

    def _predict(self, client_info, payload: MultipartPayload):
        """
        Executes request and waits for sse data
        :param payload: model input data
        :return: FlyMyAIResponse or raise an exception
        """
        return self._sse_instant(
            lambda: self._client.stream(
                method="post",
                url=client_info.prediction_path,
                timeout=_predict_timeout,
                **payload.serialize(),
                headers=client_info.authorization_headers,
            )
        )

    async def predict(
        self, payload: dict, model: Optional[str] = None, max_retries=None
    ) -> PredictionResponse:
        """
        Wrap predict method in sse.
        Retries until max_retries or self.max_retries is reached
        :param model: flymyai/bert
        :param payload: anything for model
        :param max_retries: retries
        :return: PredictionResponse(exc_history, output_data, response):
                exc_history - list of exception history during prediction
                output_data - dict with prediction output
        """
        payload = MultipartPayload(input_data=payload)
        history, response = await aretryable_callback(
            lambda: self._predict(self.amend_client_info(model), payload),
            max_retries or self.max_retries,
            FlyMyAIPredictException,
            FlyMyAIExceptionGroup,
        )
        return PredictionResponse.from_response(response, exc_history=history)

    async def predict_async_task(
        self, payload: dict, model: Optional[str] = None, max_retries=None
    ) -> AsyncPredictionTask:
        payload = MultipartPayload(input_data=payload)
        client_info = self.amend_client_info(model)
        try:
            _, response = await aretryable_callback(
                lambda: self._client.post(
                    client_info.prediction_async_path, **payload.serialize()
                ),
                max_retries or self.max_retries,
                FlyMyAIAsyncTaskException,
                FlyMyAIExceptionGroup,
            )
            response = SSEInferenceResponseFactory(response).construct()
            return self._async_prediction_task_construct(response, client_info)
        except BaseFlyMyAIException as e:
            raise FlyMyAIAsyncTaskException.from_base_exception(e)

    async def prediction_task_result(
        self, prediction_task: AsyncPredictionTask, timeout: Optional[float] = None
    ):
        prediction_id = prediction_task.prediction_id

        async def get_res():
            data_resp = await self._client.get(
                url=(
                    prediction_task.client_info or self.client_info
                ).prediction_result_path,
                params={"request_id": prediction_id},
            )
            return self._construct_task_result(data_resp)

        _, res = await aretryable_callback(
            lambda: get_res(),
            None,
            FlyMyAIAsyncTaskException,
            FlyMyAIExceptionGroup,
            timeout,
            0.5,
        )
        return res

    async def _stream(self, client_info: APIKeyClientInfo, payload: dict):
        payload = MultipartPayload(payload)
        stream_iterator = self._stream_iterator(
            client_info, payload, is_long_stream=True
        )
        decoder = SSEDecoder()
        async with stream_iterator as sse_stream:
            async for sse_partial in decoder.aiter(sse_stream.aiter_lines()):
                try:
                    response = SSEInferenceResponseFactory(
                        sse=sse_partial,
                        httpx_request=sse_stream.request,
                        httpx_response=sse_stream,
                    ).construct()
                except BaseFlyMyAIException as e:
                    raise FlyMyAIPredictException.from_base_exception(e)
                yield response

    def stream(self, payload: dict, model: Optional[str] = None, max_retries=None):
        full_client_info = self.amend_client_info(model)
        stream_iter = self._stream(full_client_info, payload)
        stream_wrapper = AsyncPredictionStream(stream_iter, self, full_client_info)
        return stream_wrapper

    @staticmethod
    async def _wrap_request(request_callback: Callable[..., Awaitable[httpx.Response]]):
        """
        Execute a request callback and return the response
        """
        response = await request_callback()
        return SSEInferenceResponseFactory(httpx_response=response).construct()

    async def close(self):
        """
        Close the client
        """
        await self._client.aclose()

    @classmethod
    async def arun_predict(cls, apikey: str, model: str, payload: dict):
        """
        Execute simple prediction out of a box
        :param model: flymyai/bert
        :param apikey: fly-...
        :param payload: {dict with prediction input}
        :return: PredictionResponse(exc_history, output_data, response)
                exc_history - list of exception history during prediction
                output_data - dict with prediction output
        """
        async with cls(apikey, model) as client:
            return await client.predict(payload)



import json
import typing

import httpx


class FlyMyAIResponse(httpx.Response):
    is_event: bool = False

    @classmethod
    def from_httpx(cls, response: httpx.Response):
        return cls(
            status_code=response.status_code,
            content=response.content,
            request=response.request,
            headers=response.headers,
        )

    def json(self, **kwargs) -> typing.Any:
        if self.content.startswith(b"data"):
            return json.loads(self.content.removeprefix(b"data"))
        elif self.content.startswith(b"event"):
            return json.loads(self.content.removeprefix(b"event"))
        else:
            return super().json(**kwargs)



from abc import abstractmethod

import httpx

from flymyai.core._response import FlyMyAIResponse
from flymyai.core._streaming import ServerSentEvent
from flymyai.core.exceptions import BaseFlyMyAIException


class ResponseFactoryException(Exception): ...


class ResponseFactory(object):
    """
    Factory for FlyMyAIResponse objects
    """

    def __init__(
        self,
        httpx_response: httpx.Response = None,
        httpx_request: httpx.Request = None,
        *_,
        **__
    ):
        if httpx_response and not httpx_request:
            self.httpx_request = httpx_response.request
        else:
            self.httpx_request = httpx_request
        self.httpx_response = httpx_response

    def _base_construct_from_httpx_response(self):
        if self.httpx_response.status_code < 400:
            return FlyMyAIResponse.from_httpx(self.httpx_response)
        else:
            raise BaseFlyMyAIException.from_response(
                FlyMyAIResponse.from_httpx(self.httpx_response)
            )

    @abstractmethod
    def construct(self):
        raise NotImplementedError



import httpx

from flymyai.core._response import FlyMyAIResponse
from flymyai.core._streaming import ServerSentEvent
from flymyai.core.exceptions import BaseFlyMyAIException
from flymyai.core.response_factory.base_response_factory import (
    ResponseFactory,
    ResponseFactoryException,
)


class SSEInferenceResponseFactory(ResponseFactory):
    def __init__(
        self,
        httpx_response: httpx.Response = None,
        httpx_request: httpx.Request = None,
        sse: ServerSentEvent = None,
    ):
        if not httpx_response and not sse:
            raise ResponseFactoryException("httpx_response and sse params required")
        super().__init__(httpx_response, httpx_request)
        self.sse = sse

    def get_sse_status_code(self):
        return self.sse.json().get(
            "status", self.httpx_response.status_code if self.httpx_response else 200
        )

    def _base_construct_from_sse(self):
        sse_status = self.get_sse_status_code()
        is_details = self.sse.json().get("details") is not None
        if is_details and sse_status == 200:
            sse_status = 599
        if sse_status < 400 and not is_details:
            response = FlyMyAIResponse(
                status_code=sse_status,
                content=self.sse.data or self.sse.event,
                request=self.httpx_request,
                headers=self.httpx_response.headers or self.sse.headers,
            )
            response.is_event = self.sse.event is not None
            return response
        else:
            raise BaseFlyMyAIException.from_response(
                FlyMyAIResponse(
                    status_code=sse_status,
                    content=self.sse.data or self.sse.event,
                    request=self.httpx_request,
                    headers=self.httpx_response.headers
                    or getattr(self.sse, "headers", {}),
                )
            )

    def construct(self):
        if self.sse:
            return self._base_construct_from_sse()
        elif self.httpx_response:
            return self._base_construct_from_httpx_response()



from typing import Optional, Iterator, TypeVar, Callable

from flymyai.core._response import FlyMyAIResponse
from flymyai.core.authorizations import APIKeyClientInfo
from flymyai.core.clients.base_client import BaseClient
from flymyai.core.exceptions import BaseFlyMyAIException
from flymyai.core.models.successful_responses import (
    StreamDetails,
    PredictionPartial,
    PredictionEvent,
)
from flymyai.core.stream_iterators.exceptions import StreamCancellationException
from flymyai.core.types.event_types import EventType


_SyncEventCallbackType = TypeVar(
    "_SyncEventCallbackType", bound=Callable[[PredictionEvent], None]
)


class PredictionStream:
    stream_details: StreamDetails
    event_callback: _SyncEventCallbackType = None
    prediction_id: str
    follow_cancelling: bool = True

    _client: BaseClient
    _client_info: APIKeyClientInfo

    def __init__(
        self,
        response_iterator: Iterator,
        client: BaseClient,
        client_info: APIKeyClientInfo,
    ):
        self.response_iterator = response_iterator
        self._client = client
        self._client_info = client_info

    def cancel(self):
        if not hasattr(self, "prediction_id"):
            raise StreamCancellationException("No prediction_id obtained!")
        return self._client.cancel_prediction(
            self.prediction_id, client_info=self._client_info
        )

    def set_on_event(self, callback: _SyncEventCallbackType):
        self.event_callback = callback

    def __iter__(self):
        return self

    def loop_iter(self):
        response_end = None
        while not response_end:
            next_resp: FlyMyAIResponse = self.response_iterator.__next__()
            if not next_resp.is_event:
                response_end = next_resp
                return response_end
            else:
                evt = PredictionEvent.from_response(next_resp)
                if evt.event_type == EventType.STREAM_ID:
                    self.prediction_id = evt.prediction_id
                if self.event_callback:
                    self.event_callback(evt)
                if self.follow_cancelling and evt.event_type == EventType.CANCELLING:
                    raise StopIteration

    def __next__(self):
        response_end = None
        try:
            response_end = self.loop_iter()
            return PredictionPartial.from_response(response_end)
        except BaseFlyMyAIException as e:
            response_end = e.response
            raise e
        except Exception as e:
            raise e
        finally:
            if not response_end:
                raise StopIteration()
            stream_details_marshalled = response_end.json().get("stream_details")
            if stream_details_marshalled:
                self.stream_details = StreamDetails.model_validate(
                    stream_details_marshalled
                )



import os
from typing import Generic, Optional, overload, AsyncIterator, Iterator, Callable
from typing import (
    TypeVar,
    Union,
)

import httpx
import pydantic

from flymyai.core.response_factory.async_task_result_factory import (
    AsyncTaskResultFactory,
)
from flymyai.core.response_factory.plain_inference_response_factory import (
    SSEInferenceResponseFactory,
)
from flymyai.core.authorizations import APIKeyClientInfo
from flymyai.core.exceptions import (
    ImproperlyConfiguredClientException,
    BaseFlyMyAIException,
    FlyMyAIAsyncTaskException,
)
from flymyai.core.models.successful_responses import (
    PredictionResponse,
    OpenAPISchemaResponse,
    PredictionPartial,
    AsyncPredictionTask,
    AsyncPredictionResponseList,
)
from flymyai.multipart import MultipartPayload

DEFAULT_RETRY_COUNT = os.getenv("FLYMYAI_MAX_RETRIES", 2)

_PossibleClients = TypeVar(
    "_PossibleClients", bound=Union[httpx.Client, httpx.AsyncClient]
)


_predict_timeout = httpx.Timeout(
    connect=int(os.getenv("FMA_CONNECT_TIMEOUT", 999999)),
    read=int(os.getenv("FMA_READ_TIMEOUT", 999999)),
    write=int(os.getenv("FMA_WRITE_TIMEOUT", 999999)),
    pool=int(os.getenv("FMA_POOL_TIMEOUT", 999999)),
)


class BaseClient(Generic[_PossibleClients]):
    """
    Base class for FlyMyAI clients
    """

    _client: _PossibleClients
    max_retries: int
    client_info: APIKeyClientInfo

    def __init__(
        self, apikey: str, model: Optional[str] = None, max_retries=DEFAULT_RETRY_COUNT
    ):
        self.client_info = APIKeyClientInfo(apikey)
        if model:
            self.client_info = self.client_info.copy_for_model(model)
        self._client = self._construct_client()
        self.max_retries = max_retries

    def amend_client_info(self, model: Optional[str] = None):
        if model:
            client_info = self.client_info.copy_for_model(model)
        else:
            client_info = self.client_info
        if not client_info.project_name or not client_info.username:
            raise ImproperlyConfiguredClientException(
                "model should be provided as <owner username>/<model>"
            )
        return client_info

    @overload
    async def predict(
        self, payload: dict, model: Optional[str] = None, max_retries=None
    ) -> PredictionResponse: ...

    @overload
    def predict(
        self, payload: dict, model: Optional[str] = None, max_retries=None
    ) -> PredictionResponse: ...

    def predict(
        self, payload: dict, model: Optional[str] = None, max_retries=None
    ) -> PredictionResponse: ...

    @overload
    async def predict_async_task(
        self, payload: dict, model: Optional[str] = None, max_retries=None
    ) -> AsyncPredictionTask: ...

    @overload
    def predict_async_task(
        self, payload: dict, model: Optional[str] = None, max_retries=None
    ) -> AsyncPredictionTask: ...

    def predict_async_task(
        self, payload: dict, model: Optional[str] = None, max_retries=None
    ) -> AsyncPredictionTask: ...

    @classmethod
    def _construct_task_result(cls, response):
        try:
            validated = AsyncTaskResultFactory(httpx_response=response).construct()
        except BaseFlyMyAIException as e:
            raise FlyMyAIAsyncTaskException.from_base_exception(e)
        try:
            return AsyncPredictionResponseList.from_response(validated, status=200)
        except pydantic.ValidationError as e:
            raise AsyncPredictionResponseList.convert_error(e) from e

    def _async_prediction_task_construct(
        self, response: httpx.Response, client_info: APIKeyClientInfo
    ) -> AsyncPredictionTask:
        async_prediction_task = AsyncPredictionTask[self.__class__].model_validate_json(
            json_data=response.content
        )
        async_prediction_task.client_info = client_info
        async_prediction_task.set_client(self)
        return async_prediction_task

    @overload
    async def prediction_task_result(
        self, prediction_task: AsyncPredictionTask, timeout: Optional[float] = None
    ): ...

    @overload
    def prediction_task_result(
        self, prediction_task: AsyncPredictionTask, timeout: Optional[float] = None
    ): ...

    def prediction_task_result(
        self, prediction_task: AsyncPredictionTask, timeout: Optional[float] = None
    ): ...

    @overload
    async def openapi_schema(
        self, model: Optional[str] = None, max_retries=None
    ) -> OpenAPISchemaResponse: ...

    @overload
    def openapi_schema(
        self, model: Optional[str] = None, max_retries=None
    ) -> OpenAPISchemaResponse: ...

    def openapi_schema(
        self, model: Optional[str] = None, max_retries=None
    ) -> OpenAPISchemaResponse: ...

    @overload
    async def stream(
        self,
        payload: dict,
        model: Optional[str] = None,
    ) -> AsyncIterator[PredictionPartial]: ...

    @overload
    def stream(
        self,
        payload: dict,
        model: Optional[str] = None,
    ) -> Iterator[PredictionPartial]: ...

    def stream(
        self,
        payload: dict,
        model: Optional[str] = None,
    ): ...

    def _stream_iterator(
        self, client_info, payload: MultipartPayload, is_long_stream: bool
    ) -> Union[Iterator[httpx.Response], AsyncIterator[httpx.Response]]:
        return self._client.stream(
            method="post",
            url=(
                client_info.prediction_path
                if not is_long_stream
                else client_info.prediction_stream_path
            ),
            **payload.serialize(),
            timeout=_predict_timeout,
            headers=client_info.authorization_headers,
            follow_redirects=True,
        )

    @staticmethod
    def _wrap_request(request_callback: Callable):
        response = request_callback()
        return SSEInferenceResponseFactory(httpx_response=response).construct()

    def is_closed(self) -> bool:
        return self._client.is_closed

    def close(self) -> None:
        """
        Close the underlying HTTPX client.

        The client will *not* be usable after this.
        """
        # If an error is thrown while constructing a client, self._client
        # may not be present
        if hasattr(self, "_client"):
            self._client.close()

    def _construct_client(self):
        raise NotImplemented

    @overload
    async def cancel_prediction(
        self,
        prediction_id: str,
        model: Optional[str] = None,
        client_info: APIKeyClientInfo = None,
    ): ...

    @overload
    def cancel_prediction(
        self,
        prediction_id: str,
        model: Optional[str] = None,
        client_info: APIKeyClientInfo = None,
    ): ...

    def cancel_prediction(
        self,
        prediction_id: str,
        model: Optional[str] = None,
        client_info: APIKeyClientInfo = None,
    ): ...



import os
from typing import Callable, Iterator, Optional

import httpx

from flymyai.core._streaming import SSEDecoder
from flymyai.core.authorizations import APIKeyClientInfo
from flymyai.core.clients.base_client import BaseClient, _predict_timeout
from flymyai.core.exceptions import (
    BaseFlyMyAIException,
    FlyMyAIOpenAPIException,
    FlyMyAIPredictException,
    FlyMyAIExceptionGroup,
    FlyMyAIAsyncTaskException,
)
from flymyai.core.models.successful_responses import (
    PredictionResponse,
    OpenAPISchemaResponse,
    AsyncPredictionTask,
)
from flymyai.core.response_factory.plain_inference_response_factory import (
    SSEInferenceResponseFactory,
)
from flymyai.core.stream_iterators.PredictionStream import PredictionStream
from flymyai.multipart import MultipartPayload
from flymyai.utils.utils import retryable_callback


class BaseSyncClient(BaseClient[httpx.Client]):
    def _construct_client(self):
        return httpx.Client(
            http2=True,
            headers=self.client_info.authorization_headers,
            base_url=os.getenv("FLYMYAI_DSN", "https://api.flymy.ai/"),
            timeout=_predict_timeout,
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._client.close()

    @classmethod
    def _sse_instant(cls, stream_iter_func: Callable[[], Iterator[httpx.Response]]):
        """
        Fetch sse response on prediction
        :param stream_iter_func: context manager with underlying stream
        :return: FlyMyAIResponse
        """
        with stream_iter_func() as stream:
            stream: httpx.Response
            response = SSEInferenceResponseFactory(
                sse=next(SSEDecoder().iter(stream.iter_lines())),
                httpx_request=stream.request,
                httpx_response=stream,
            ).construct()
            return response

    def _predict(self, payload: MultipartPayload, client_info: APIKeyClientInfo):
        """
        Wrap predict method in sse
        """

        try:
            return self._sse_instant(
                lambda: self._stream_iterator(client_info, payload, False)
            )
        except BaseFlyMyAIException as e:
            raise FlyMyAIPredictException.from_base_exception(e)

    def predict(self, payload: dict, model: Optional[str] = None, max_retries=None):
        """
        Wrap predict method in sse.
        Retries until max_retries or self.max_retries is reached
        :param model: flymyai/bert | None, If none - get self.client_info.<username/project_name>
        :param payload: anything for model
        :param max_retries: retries
        :return: PredictionResponse(exc_history, output_data, response):
                exc_history - list of exception history during prediction
                output_data - dict with prediction output
        """

        payload = MultipartPayload(payload)
        history, response = retryable_callback(
            lambda: self._predict(payload, self.amend_client_info(model)),
            max_retries or self.max_retries,
            FlyMyAIPredictException,
            FlyMyAIExceptionGroup,
        )
        return PredictionResponse.from_response(response, exc_history=history)

    def predict_async_task(
        self, payload: dict, model: Optional[str] = None, max_retries=None
    ):
        payload = MultipartPayload(input_data=payload)
        client_info = self.amend_client_info(model)
        try:
            _, response = retryable_callback(
                lambda: self._client.post(
                    client_info.prediction_async_path, **payload.serialize()
                ),
                max_retries or self.max_retries,
                FlyMyAIAsyncTaskException,
                FlyMyAIExceptionGroup,
            )
            response = SSEInferenceResponseFactory(response).construct()
            return self._async_prediction_task_construct(response, client_info)
        except BaseFlyMyAIException as e:
            raise FlyMyAIAsyncTaskException.from_base_exception(e)

    def prediction_task_result(
        self, prediction_task: AsyncPredictionTask, timeout: Optional[float] = None
    ):
        prediction_id = prediction_task.prediction_id

        def get_res():
            resp = self._client.get(
                url=(
                    prediction_task.client_info or self.client_info
                ).prediction_result_path,
                params={"request_id": prediction_id},
            )
            return self._construct_task_result(resp)

        _, res = retryable_callback(
            lambda: get_res(),
            None,
            FlyMyAIAsyncTaskException,
            FlyMyAIExceptionGroup,
            timeout,
            0.5,
        )

        return res

    def _stream(self, client_info: APIKeyClientInfo, payload: dict):
        payload = MultipartPayload(payload)
        response_iterator = self._stream_iterator(
            client_info, payload, is_long_stream=True
        )
        decoder = SSEDecoder()
        with response_iterator as sse_stream:
            for sse_partial in decoder.iter(sse_stream.iter_lines()):
                try:
                    response = SSEInferenceResponseFactory(
                        sse=sse_partial,
                        httpx_request=sse_stream.request,
                        httpx_response=sse_stream,
                    ).construct()
                except BaseFlyMyAIException as e:
                    raise FlyMyAIPredictException.from_base_exception(e)
                yield response

    def stream(self, payload: dict, model: Optional[str] = None):
        full_client_info = self.amend_client_info(model)
        stream_iter = self._stream(full_client_info, payload)
        stream_wrapper = PredictionStream(stream_iter, self, full_client_info)
        return stream_wrapper

    def _openapi_schema(self, client_info: APIKeyClientInfo):
        """
        OpenAPI request for the current project, wrapped in executor-method (using HTTP/1)
        :return:
        """
        try:
            return self._wrap_request(
                lambda: self._client.get(
                    client_info.openapi_schema_path,
                    headers=client_info.authorization_headers,
                )
            )
        except BaseFlyMyAIException as e:
            raise FlyMyAIOpenAPIException.from_response(e.response)

    def cancel_prediction(
        self,
        prediction_id: str,
        model: Optional[str] = None,
        client_info: APIKeyClientInfo = None,
    ):
        if client_info:
            full_client_info = client_info
        else:
            full_client_info = self.amend_client_info(model)
        response = self._client.patch(
            url=full_client_info.prediction_cancel_path,
            json={"infer_id": prediction_id},
        )
        return SSEInferenceResponseFactory(
            httpx_response=response, httpx_request=response.request
        ).construct()

    def openapi_schema(self, model: Optional[str] = None, max_retries=None):
        """
        :param model: flymyai/bert
        :param max_retries: retries before give up
        :return:
        :return: OpenAPISchemaResponse(exc_history, openapi_schema, response):
                exc_history - dict with exceptions;
                openapi_schema - dict with openapi;
        """
        history, response = retryable_callback(
            lambda: self._openapi_schema(client_info=self.amend_client_info(model)),
            max_retries or self.max_retries,
            FlyMyAIPredictException,
            FlyMyAIExceptionGroup,
        )
        return OpenAPISchemaResponse.from_response(
            exc_history=history, openapi_schema=response.json(), response=response
        )

    @classmethod
    def run_predict(cls, apikey: str, model: str, payload: dict):
        """
        :param apikey: fly-...
        :param model:  flymyai/bert
        :param payload: jsonable / multipart/form-data available data
        :return: PredictionResponse(exc_history, output_data, response):
                exc_history - list of exception history during prediction;
                output_data - dict with prediction output;
        """
        with cls(apikey, model) as client:
            return client.predict(payload)



from flymyai.core.response_factory.base_response_factory import ResponseFactory


class MaybeNotExistent(Exception): ...


class AsyncTaskResultFactory(ResponseFactory):
    def construct(self):
        return self._base_construct_from_httpx_response()



from typing import Optional, Generic, TypeVar, List, TypedDict, Union, Awaitable

import pydantic
from pydantic import PrivateAttr, model_validator, Field
from pydantic_core._pydantic_core import PydanticCustomError
from typing_extensions import Self

from flymyai.core._response import FlyMyAIResponse
from flymyai.core.authorizations import APIKeyClientInfo
from flymyai.core.exceptions import BaseFlyMyAIException, FlyMyAIExceptionGroup
from flymyai.core.models.base import ResponseLike
from flymyai.core.types.event_types import EventType

_ClientT = TypeVar("_ClientT", bound="BaseClient")


class BaseFromServer(pydantic.BaseModel):
    _response: Optional[FlyMyAIResponse] = PrivateAttr(default=None)

    @property
    def response(self):
        return self._response

    @classmethod
    def from_response(cls, response: FlyMyAIResponse, **kwargs):
        status_code = kwargs.pop("status", response.status_code)
        response_json = response.json()
        response_json["status"] = response_json.get("status", status_code)
        ctx = kwargs.pop("context", None)
        self = cls.model_validate(dict(**response_json, **kwargs), context=ctx)
        self._response = response
        return self


class BasePredictionResponse(BaseFromServer):
    exc_history: Optional[list] = Field(default_factory=list)
    output_data: dict


class PredictionResponse(BasePredictionResponse):
    """
    Prediction response from FlyMyAI
    """

    status: int
    inference_time: Optional[float] = None

    @property
    def response(self):
        return self._response


class AsyncPredictionResponse(BasePredictionResponse):
    infer_details: dict
    output_data: dict = Field(validation_alias="response")

    @property
    def status(self) -> int:
        return self.infer_details.get("status", 200)

    @model_validator(mode="after")
    def validate_inference_details(self, ctx):
        self._response = ctx.context.get("_response")
        if (status := self.infer_details.get("status", 200)) != 200:
            raise PydanticCustomError(
                "inference_response_error",
                "Inference details contains incorrect status: {failure_status}",
                dict(failure_status=status, instance=self),
            )
        return self


class AsyncPredictionResponseList(BaseFromServer):
    inference_responses: List[AsyncPredictionResponse]

    @classmethod
    def from_response(cls, response: FlyMyAIResponse, **kwargs):
        result = super().from_response(
            response, context={"_response": response}, **kwargs
        )
        return result

    class _ErrorCTX(TypedDict):
        failure_status: int
        instance: Self

    @classmethod
    def convert_error(cls, e: pydantic.ValidationError) -> FlyMyAIExceptionGroup:
        errors = []
        error_data = e.errors(include_input=False)
        for error in error_data:
            err_t = error.get("type")
            ctx: Optional[cls._ErrorCTX] = error.get("ctx", {})
            if err_t == "inference_response_error":
                if not ctx:
                    raise KeyError("Inference response pydantic error should have ctx!")
                errors.append(
                    BaseFlyMyAIException.from_response(
                        ResponseLike(
                            status_code=ctx["failure_status"],
                            url=getattr(ctx["instance"], "_response").url,
                            content=ctx["instance"].model_dump_json().encode(),
                        )
                    )
                )
        if len(errors) != error_data:
            errors.append(e)
        exc_group = FlyMyAIExceptionGroup(errors)
        return exc_group


class AsyncPredictionTask(BaseFromServer, Generic[_ClientT]):
    _affiliated_client: Optional[_ClientT] = PrivateAttr(default=None)
    _client_info: APIKeyClientInfo = PrivateAttr(default=None)

    prediction_id: str

    def result(
        self, timeout=None
    ) -> Union[AsyncPredictionResponseList, Awaitable[AsyncPredictionResponseList]]:
        return self._affiliated_client.prediction_task_result(self, timeout=timeout)

    @property
    def client_info(self) -> APIKeyClientInfo:
        return self._client_info

    @client_info.setter
    def client_info(self, v: APIKeyClientInfo):
        self._client_info = v

    def set_client(self, client: _ClientT):
        self._affiliated_client = client


class OpenAPISchemaResponse(BaseFromServer):
    """
    OpenAPI schema for the current project. Use it to construct your own schema
    """

    exc_history: Optional[list]
    openapi_schema: dict
    status: int


class PredictionPartial(BaseFromServer):
    status: int
    output_data: Optional[dict] = None

    _response: FlyMyAIResponse = PrivateAttr()


class PredictionEvent(BaseFromServer):
    status: int
    event_type: EventType

    prediction_id: Optional[str] = None  # EventType.STREAM_ID


class StreamDetails(pydantic.BaseModel):
    input_tokens: Optional[int] = None
    output_tokens: Optional[int] = None
    size_in_billions: Optional[float] = pydantic.Field(
        default=None, alias="model_size_in_billions"
    )



import dataclasses
import json
from typing import Union

import httpx

from flymyai.core.models.base import ResponseLike


@dataclasses.dataclass
class Base4xxResponse(ResponseLike):
    """
    Base class for all 4xx
    """

    status_code: int
    url: httpx.URL
    content: bytes

    requires_retry: bool = False

    @classmethod
    def from_response(cls, response: Union[httpx.Response, ResponseLike]):
        return cls(response.status_code, response.url, response.content)


@dataclasses.dataclass
class FlyMyAI400Response(Base4xxResponse):
    """
    400 response
    requires_retry = False means we do not resend this request
    """

    requires_retry: bool = False

    def to_msg(self):
        return f"""
            Bad request happened: {self.content.decode()}
        """


@dataclasses.dataclass
class FlyMyAI401Response(Base4xxResponse):
    """
    401 response
    requires_retry = False means we do not resend this request
    """

    requires_retry: bool = False

    def to_msg(self):
        return f"""
            Authentication error: verify your credentials!
        """


@dataclasses.dataclass
class FlyMyAI421Response(Base4xxResponse):
    requires_retry = False

    def to_msg(self):
        jsoned = json.loads(self.content)
        msg = super().to_msg()
        if detail := jsoned.get("detail"):
            msg += f"\nDetail: {detail}"
        return msg


@dataclasses.dataclass
class FlyMyAI422Response(Base4xxResponse):
    """
    422 response
    requires_retry = False means we do not resend this request
    """

    requires_retry: bool = False

    def to_msg(self):
        jsoned = json.loads(self.content)
        msg = super().to_msg()
        if detail := jsoned.get("detail"):
            msg += f"Details: {detail}"
        return msg


@dataclasses.dataclass
class FlyMyAI425Response(Base4xxResponse):
    requires_retry: bool = True



import asyncio
import threading
import time
from typing import Callable, Awaitable, Type, Optional

import httpx

from flymyai.core.exceptions import RetryTimeoutExceededException


def retryable_callback(
    cb: Callable,
    retries: Optional[int],
    append_on_exception_cls: Type[Exception],
    exception_group_cls: Type[Exception],
    timeout_seconds: Optional[float] = None,
    await_treshold: Optional[float] = None,
):
    """
    Decorator to retry a function
    """

    should_stop = False
    result_container = None
    exc_container = None

    def wrapper():
        nonlocal should_stop, result_container, exc_container
        retries_history = []
        r = 0
        while r != retries:
            try:
                res = cb()
                result_container = retries_history, res
                return
            except append_on_exception_cls as e:
                retries_history.append(e)
                if e.requires_retry and not should_stop:
                    time.sleep(await_treshold or 0)
                    r += 1
                    continue
                else:
                    exc_container = exception_group_cls(retries_history)
                    return
            except exception_group_cls as e1:
                exc_container = e1
                return
            except Exception as e2:
                exc_container = e2
                return
        exception_gr = exception_group_cls(retries_history)
        exc_container = exception_gr

    waiting_thread = threading.Thread(target=wrapper)
    waiting_thread.start()
    if timeout_seconds is not None:
        timeout_seconds += 0.01
    waiting_thread.join(timeout=timeout_seconds)
    should_stop = True
    if not result_container and not exc_container:
        raise RetryTimeoutExceededException()
    if exc_container:
        raise exc_container
    return result_container


async def aretryable_callback(
    cb: Callable[..., Awaitable[httpx.Response]],
    retries,
    append_on_exception_cls: Type[Exception],
    exception_group_cls: Type[Exception],
    timeout_seconds: Optional[float] = None,
    await_treshold: Optional[float] = None,
):
    """
    Decorator to retry a function
    """
    if timeout_seconds is not None:
        timeout_seconds += 0.01

    async def wrapper():
        retries_history = []
        r = 0
        while r != retries:
            try:
                res = await cb()
                return retries_history, res
            except append_on_exception_cls as e1:
                retries_history.append(e1)
                if e1.requires_retry:
                    await asyncio.sleep(await_treshold or 0)
                    r += 1
                    continue
                else:
                    raise exception_group_cls(retries_history)
            except Exception as e2:
                raise e2
        else:
            exception_gr = exception_group_cls(retries_history)
            raise exception_gr

    try:
        return await asyncio.wait_for(wrapper(), timeout_seconds)
    except asyncio.TimeoutError as e:
        raise RetryTimeoutExceededException() from e



import asyncio
from typing import AsyncIterator, TypeVar, Callable, Union, Awaitable

from flymyai.core._response import FlyMyAIResponse
from flymyai.core.authorizations import APIKeyClientInfo
from flymyai.core.clients.base_client import BaseClient
from flymyai.core.exceptions import BaseFlyMyAIException
from flymyai.core.models.successful_responses import (
    StreamDetails,
    PredictionPartial,
    PredictionEvent,
)
from flymyai.core.stream_iterators.exceptions import StreamCancellationException
from flymyai.core.types.event_types import EventType


_AsyncEventCallbackType = TypeVar(
    "_AsyncEventCallbackType",
    bound=Union[
        Callable[[PredictionEvent], None], Callable[[PredictionEvent], Awaitable[None]]
    ],
)


class AsyncPredictionStream:
    stream_details: StreamDetails

    event_callback: _AsyncEventCallbackType = None

    prediction_id: str

    follow_cancelling: bool = True

    _client: BaseClient
    _client_info: APIKeyClientInfo

    def __init__(
        self,
        response_iterator: AsyncIterator,
        client: BaseClient,
        client_info: APIKeyClientInfo,
    ):
        self.response_iterator = response_iterator
        self._client = client
        self._client_info = client_info

    async def cancel(self):
        if not hasattr(self, "prediction_id"):
            raise StreamCancellationException("No prediction_id obtained!")
        return await self._client.cancel_prediction(
            self.prediction_id, client_info=self._client_info
        )

    def __aiter__(self):
        return self

    def set_on_event(self, callback_or_coro: _AsyncEventCallbackType):
        self.event_callback = callback_or_coro

    async def loop_iter(self):
        response_end = None
        while not response_end:
            next_resp: FlyMyAIResponse = await self.response_iterator.__anext__()
            if not next_resp.is_event:
                response_end = next_resp
                return response_end
            else:
                evt = PredictionEvent.from_response(next_resp)
                if not self.event_callback:
                    pass
                else:
                    coro_or_res = self.event_callback(evt)
                    if asyncio.iscoroutine(coro_or_res):
                        asyncio.run_coroutine_threadsafe(
                            coro_or_res, asyncio.get_event_loop()
                        )
                if self.follow_cancelling and evt.event_type == EventType.CANCELLING:
                    raise StopAsyncIteration
                if evt.event_type == EventType.STREAM_ID:
                    self.prediction_id = evt.prediction_id

    async def __anext__(self):
        response_end = None
        try:
            response_end = await self.loop_iter()
            return PredictionPartial.from_response(response_end)
        except BaseFlyMyAIException as e:
            response_end = e.response
            raise e
        except Exception as e:
            raise e
        finally:
            if not response_end:
                raise StopAsyncIteration()
            stream_details_marshalled = response_end.json().get("stream_details")
            if stream_details_marshalled:
                self.stream_details = StreamDetails.model_validate(
                    stream_details_marshalled
                )
